from flask import jsonify, request
from loguru import logger

from ..app import app
from ..interface import interface


@app.route("/hw_proxy/default_printer_action", methods=["POST", "PUT"])
@logger.catch
def default_printer_action():
    logger.trace("default_printer_action()")
    data = request.json.get("params", {}).get("data", {})

    data_copy = data.copy()
    if data_copy.get("action") == "print_receipt":
        data_copy.pop("receipt")
    interface.log_data(data_copy, "printer", "DEBUG")

    if data.get("action") == "print_receipt":
        receipt = data["receipt"]
        interface.log_image(receipt, "DEBUG")
        interface.device_printer_task_print(receipt)

    elif data.get("action") == "cashbox":
        interface.device_printer_task_open_cashbox()

    else:
        raise Exception(f"Incorrect action value: '{data.get('action')}'")

    return jsonify(jsonrpc="2.0", result=True)



from flask_babel import gettext as _
from loguru import logger

from .models.device_abstract_with_queue import DeviceAbstractWithQueue
from .tools.serial_code import FF, US


class DeviceDisplay(DeviceAbstractWithQueue):
    use_serial = True
    device_type = "display"

    @logger.catch
    def _process_task_serial(self, serial, task):
        lines_ascii = self._convert_task_to_ascii_lines(task)

        self._clear_screen(serial)

        for row, text in enumerate(lines_ascii, 1):
            self._position_cursor(serial, 1, row)
            self._send_message(serial, [text])

    def _convert_task_to_ascii_lines(self, task):
        lines_ascii = []

        action = task.data["action"]

        if action == "display_lines":
            for line in task.data["text_lines"]:
                lines_ascii.append(line)

        elif action in [
            "add_line",
            "update_quantity",
            "update_unit_price",
            "update_discount",
        ]:
            # First Line
            left_part = task.data.get("product_name")
            right_part = self._prepare_percent(task.data.get("discount"))
            line_1 = self._prepare_line(left_part, right_part)

            # Second Line Line
            left_part = (
                self._prepare_qty(task.data.get("quantity"))
                + "×"
                + self._prepare_price(task.data.get("unit_price"))
            )
            right_part = self._prepare_price(task.data.get("total"))
            line_2 = self._prepare_line(left_part, right_part)

            lines_ascii = [line_1, line_2]

        elif action == "remove_line":
            lines_ascii = [
                _("Deleting Line ..."),
                self._prepare_line(task.data.get("product_name")),
            ]

        elif action == "payment_balance":
            total = task.data.get("total")
            total_paid = task.data.get("total_paid")
            total_due = task.data.get("total_due")
            total_to_pay = total - total_paid
            line_1 = self._prepare_line(_("Total:"), self._prepare_price(total))
            line_2 = ""
            if total_paid != 0.0:
                if total_to_pay > 0:
                    line_2 = self._prepare_line(
                        _("To Pay:"), self._prepare_price(total_to_pay)
                    )
                elif total_due < 0:
                    line_2 = self._prepare_line(
                        _("Change:"), self._prepare_price(-total_due)
                    )
            lines_ascii = [line_1, line_2]

        else:
            raise NotImplementedError(
                f"{action} is not a valid action for the Display device."
            )

        return lines_ascii

    def _clear_screen(self, serial):
        self._send_message(serial, [FF], "Clear Screen")

    def _position_cursor(self, serial, col=1, row=1):
        self._send_message(
            serial, [US, chr(36), chr(col), chr(row)], f"Move to col {col} / row {row}"
        )

    def _send_message(self, serial, message_list, action_description=""):
        message_list = [x for x in message_list if x]
        if not message_list:
            return
        message = b"".join([x.encode("cp858", errors="ignore") for x in message_list])
        if action_description:
            logger.debug(
                f"Display: {action_description} with the message {message} ..."
            )

        serial.write(message)

    def _display_float(self, float_repr, reduce_txt=True, suffix=""):
        if reduce_txt:
            while float_repr:
                if float_repr[-1:] == "0":
                    float_repr = float_repr[:-1]
                else:
                    break
            if float_repr[-1:] == ".":
                float_repr = float_repr[:-1]
        return float_repr.replace(".", ",") + suffix

    def _prepare_percent(self, percent):
        try:
            percent = float(percent)
        except ValueError:
            return ""
        except TypeError:
            return ""
        if not percent:
            return ""
        return self._display_float("-{:.1f}".format(percent), suffix="%")

    def _prepare_qty(self, quantity):
        try:
            quantity = float(quantity)
        except ValueError:
            quantity = 0.0
        return self._display_float("{:.3f}".format(quantity))

    def _prepare_price(self, price):
        try:
            price = float(price)
        except ValueError:
            price = 0.0
        return self._display_float("{:.2f}".format(price), reduce_txt=False, suffix="€")

    def _prepare_line(self, left_part, right_part=""):
        result = left_part[:20].ljust(20, " ")
        if right_part:
            right_part = right_part[-20:]
            result = result[0 : (20 - 1 - len(right_part))]
            result += " " + right_part

        return result



from .device_display import DeviceDisplay  # noqa: F401
from .device_payment import DevicePayment  # noqa: F401
from .device_printer import DevicePrinter  # noqa: F401
from .device_scale import DeviceScale  # noqa: F401



import curses.ascii

STX = chr(curses.ascii.STX)
ETX = chr(curses.ascii.ETX)
EOT = chr(curses.ascii.EOT)
ENQ = chr(curses.ascii.ENQ)
ACK = chr(curses.ascii.ACK)
NAK = chr(curses.ascii.NAK)
ESC = chr(curses.ascii.ESC)

FF = chr(curses.ascii.FF)
US = chr(curses.ascii.US)



import simplejson
from flask import jsonify, request
from loguru import logger

from ..app import app
from ..interface import interface


@app.route("/hw_proxy/display_show", methods=["POST", "PUT"])
@logger.catch
def display_show():
    logger.trace("display_show()")
    data = request.json.get("params", {}).get("data", {})
    interface.log_data(data, "display", "DEBUG")

    if type(data) is str:
        data = simplejson.loads(data)

    interface.device_display_task_show(data)
    return jsonify(jsonrpc="2.0", result=True)



import serial.tools.list_ports
from flask_babel import gettext as _
from loguru import logger
from serial import Serial

from ...usb_device import usb_device_tools


class DeviceAbstract:
    has_queue = False
    use_serial = False
    device_type = False

    def __init__(self, interface, options):
        self._interface = interface
        self._usb_device = False
        self.usb_device_name = False
        self.terminal_file = False
        self._usb_image_name = False
        self._options = options

    def get_option(self, option_name, option_type, required=False):
        option = self._options.get(option_name, False)
        if required and not option:
            raise Exception(
                f"{option_name} is required for the device {self.device_type}."
            )
        return option_type(option)

    @property
    def name(self):
        if self.device_type == "display":
            return _("Display")
        elif self.device_type == "printer":
            return _("Printer")
        elif self.device_type == "payment":
            return _("Payment Terminal")
        elif self.device_type == "scale":
            return _("Scale")
        return "N/C"

    @property
    def usb_vendor_product_code(self):
        return usb_device_tools.get_device_id_vendor_id_product(self._usb_device)

    @property
    def usb_serial_number(self):
        return usb_device_tools.get_device_serial_number(self._usb_device)

    @property
    def usb_manufacturer(self):
        return usb_device_tools.get_device_manufacturer(self._usb_device)

    @property
    def usb_product_name(self):
        return usb_device_tools.get_device_product(self._usb_device)

    @property
    def is_connected(self):
        return bool(self._usb_device)

    @property
    def disconnections(self):
        return self._interface.disconnections.get(self.usb_vendor_product_code, [])

    @property
    def disconnections_qty(self):
        return len(self.disconnections)

    @property
    def usb_image_name(self):
        return self._usb_image_name or f"{self.device_type}.png"

    # ###########################
    # Device Section
    # ###########################
    def set_usb_device(self, usb_device, extra_info):
        self._usb_device = usb_device
        self.usb_device_name = extra_info.get("name", False)
        self._usb_image_name = extra_info.get("image")
        connected_comports = [x for x in serial.tools.list_ports.comports()]
        for port in connected_comports:
            if (port.vid, port.pid) == (
                usb_device.idVendor,
                usb_device.idProduct,
            ):
                self.terminal_file = port.device
                self._logger("DEBUG", f"Terminal File found: {self.terminal_file}.")
                break
        if not self.terminal_file:
            self._logger("DEBUG", "Terminal File not found.")

    def remove_usb_device(self):
        self._usb_device = False
        self.usb_device_name = False
        self._usb_image_name = False
        self.terminal_file = False

    def _get_serial(self):
        if not self.terminal_file:
            raise Exception(
                "Unable to open a Serial connexion,"
                " because terminal_file is not defined."
            )
        return Serial(self.terminal_file, 9600, timeout=0.05)

    # ###########################
    # Loggging Section
    # ###########################
    def _logger(self, level, message):
        if self._usb_device:
            extra_info = f" ({self.usb_device_name} - {self.usb_vendor_product_code})"
        else:
            extra_info = ""
        logger.log(level, f"Device '{self.device_type}'{extra_info}: {message}")



from flask import jsonify, make_response
from loguru import logger

from ..app import app
from ..interface import interface


@app.route("/hw_proxy/hello")
@logger.catch
def hello():
    """
    hw_proxy/hello is called when the Point of Sale is launched,
    and when user clicks on the device icon.

    If the call to hello fails, Iot Box is considered offline
    and no other calls will be done."""
    logger.trace("hello()")
    return make_response("ping")


@app.route("/hw_proxy/handshake", methods=["POST"])
@logger.catch
def handshake():
    """hw_proxy/handshake is done just after 'hello' call,
    if the hello call succeeded.

    If the call to handshake fails, an error is raised
    in the Point of sale, and no other calls will be done.
    """
    logger.trace("handshake()")
    return jsonify(jsonrpc="2.0", result=True)


@app.route("/hw_proxy/status_json", methods=["POST", "PUT"])
@logger.catch
def status_json():
    """
    hw_proxy/status_json is called every 5 seconds by Odoo,
    if hello / handshake was correctly answered.

    it returns a dictionary with the following structure.
        {
            'DEVICE_TYPE_1': {'status': 'STATUS_1'},
            'DEVICE_TYPE_X': {'status': 'STATUS_X'},
        }

    If the call to status_json fails, an error is raised
    every each 5 seconds, that is quite annoying for the end user.
    """
    logger.trace("status_json()")
    status = {}

    for device_type in interface.device_types:
        device = interface.get_device(device_type)
        if device.is_connected:
            status[device_type] = {
                "status": "connected",
                "vendor_product_code": device.usb_vendor_product_code,
                "serial_number": device.usb_serial_number,
                "manufacturer": device.usb_manufacturer,
                "product_name": device.usb_product_name,
                "device_name": device.usb_device_name,
            }
        else:
            status[device_type] = {"status": "disconnected"}

    return jsonify(jsonrpc="2.0", result=status)



from . import usb_device_list


def get_device_id_vendor_id_product(device):
    if device:
        return (
            "{0:04x}".format(device.idVendor) + ":" + "{0:04x}".format(device.idProduct)
        )
    else:
        return "N/C"


def get_device_manufacturer(device):
    try:
        return device.manufacturer
    except ValueError:
        return ""


def get_device_serial_number(device):
    try:
        return device.serial_number and device.serial_number or ""
    except ValueError:
        return ""


def get_device_product(device):
    try:
        return device.product
    except ValueError:
        return ""


def get_device_type(device):
    try:
        (device_type, _extra_info) = usb_device_list.get_device_type(
            device.idVendor, device.idProduct
        )
        return device_type
    except ValueError:
        return False



import base64
import time
from datetime import datetime
from io import BytesIO
from threading import Thread

import usb
from loguru import logger
from PIL import Image

from .devices import DeviceDisplay, DevicePayment, DevicePrinter, DeviceScale
from .usb_device import usb_device_list, usb_device_tools


class Interface(Thread):
    device_types = ["display", "printer", "payment", "scale"]

    def initialize(self, refresh_devices_delay=1, options={}, log_folder=False):
        self.disconnections = {}
        self.start_date = datetime.now()
        self.usb_devices = []
        self.refresh_devices_delay = refresh_devices_delay
        self.log_folder = log_folder
        self.usb_device_display = DeviceDisplay(
            self, options.get("display", {}), delay=0.2, max_queue_size=10
        )
        self.usb_device_printer = DevicePrinter(
            self, options.get("printer", {}), delay=1, max_queue_size=3
        )

        self.usb_device_payment = DevicePayment(self, options.get("payment", {}))

        self.usb_device_scale = DeviceScale(self, options.get("scale", {}))

        self.refresh_usb_devices()

        for device in [self.get_device(x) for x in self.device_types]:
            if device.has_queue:
                device.start()

        self.start()

    # ###########################
    # Public Implemented Devices Section
    # ###########################

    def device_display_task_show(self, data):
        self.usb_device_display.add_task(data)

    def device_printer_task_print(self, receipt):
        self.usb_device_printer.add_task(("print", receipt))

    def device_printer_task_open_cashbox(self):
        self.usb_device_printer.add_task(("open_cashbox", False))

    def device_scale_read_weight(self, unit_price, tare):
        return self.usb_device_scale.read_weight(unit_price, tare)

    def device_payment_push_amount(self, amount, currency_iso):
        return self.usb_device_payment.push_amount(amount, currency_iso)

    # ###########################
    # Generic Device Section
    # ###########################

    def get_device(self, device_type):
        return getattr(self, f"usb_device_{device_type}")

    @logger.catch
    def refresh_usb_devices(self):
        logger.debug(
            "Refreshing USB Devices List ..."
            f"(Next refresh in {self.refresh_devices_delay} second(s).)"
        )
        new_devices = [device for device in usb.core.find(find_all=True)]

        # Detect new devices
        for device in new_devices:
            if device not in self.usb_devices:
                self._hook_usb_device_new(device)

        # Handle device removal
        for device in self.usb_devices:
            if device not in new_devices:
                self._hook_usb_device_removed(device)

    def _hook_usb_device_new(self, device):
        """Handle here things to be done, when a new device is detected"""
        (device_type, extra_info) = usb_device_list.get_device_type(
            device.idVendor, device.idProduct
        )
        if (
            usb_device_tools.get_device_id_vendor_id_product(device)
            not in self.disconnections.keys()
        ):
            self.disconnections[
                usb_device_tools.get_device_id_vendor_id_product(device)
            ] = []

        logger.log(
            device_type and "SUCCESS" or "INFO",
            f"Found new USB device."
            f" Type '{device_type or 'Unknown'}'."
            f" Name '{device_type and extra_info['name'] or 'Unknown'}'."
            " Code:"
            f" {usb_device_tools.get_device_id_vendor_id_product(device)} ;"
            f" Tecnical Name: {usb_device_tools.get_device_product(device)}",
        )
        if device_type:
            getattr(self, f"usb_device_{device_type}").set_usb_device(
                device, extra_info
            )
        self.usb_devices.append(device)

    def _hook_usb_device_removed(self, device, error=False):
        """Handle here cleanup to be done, when a removal of
        device is detected"""
        (device_type, _extra_info) = usb_device_list.get_device_type(
            device.idVendor, device.idProduct
        )
        self.disconnections[
            usb_device_tools.get_device_id_vendor_id_product(device)
        ].append({"date": datetime.now(), "error": error})
        logger.info(
            f"Removing USB device."
            " Code:"
            f" {usb_device_tools.get_device_id_vendor_id_product(device)} ;"
            f" Name: {usb_device_tools.get_device_product(device)}"
        )
        if device_type:
            getattr(self, f"usb_device_{device_type}").remove_usb_device()
        self.usb_devices.remove(device)

    # ###########################
    # Thread Section
    # ###########################

    def run(self):
        while True:
            time.sleep(self.refresh_devices_delay)
            self.refresh_usb_devices()

    # ###########################
    # Log Section
    # ###########################
    def log_image(self, base64_image, log_level):
        if logger._core.min_level <= logger.level(log_level).no:
            try:
                image = Image.open(BytesIO(base64.b64decode(base64_image)))
                dt = datetime.now().isoformat("_").replace(":", "-").replace(".", "-")
                filepath = self.log_folder / f"odoo-driver__print__{dt}.{image.format}"
                image.save(filepath)
                logger.debug(f"Image saved into {filepath}")
            except Exception as e:
                logger.error(f"Unable to log image. {e}")

    def log_data(self, data, device_type, log_level):
        if logger._core.min_level <= logger.level(log_level).no:
            logger.debug(f"Data received on device '{device_type}': {data}")


interface = Interface()



import os
import platform

import psutil
from flask import render_template, send_from_directory
from flask_babel import gettext as _
from loguru import logger

from .. import __version__
from ..app import app
from ..interface import interface
from ..usb_device import usb_device_tools


@app.route("/")
@app.route("/home.html")
@logger.catch
def home():
    logger.trace("home()")
    return render_template(
        "home.html.jinja",
        interface=interface,
        usb_device_tools=usb_device_tools,
    )


@app.route("/favicon.ico")
@logger.catch
def favicon():
    return send_from_directory(
        os.path.join(app.root_path, "static", "img"),
        "favicon.png",
        mimetype="image/png",
    )


@app.route("/usb_devices.html", methods=["GET"])
@logger.catch
def usb_devices():
    logger.trace("usb_devices()")
    return render_template(
        "usb_devices.html.jinja",
        interface=interface,
        usb_device_tools=usb_device_tools,
    )


@app.route("/system.html")
@logger.catch
def system():
    logger.trace("system()")
    system_info = [
        (_("Operating System"), platform.system()),
        (_("Kernel Release"), platform.release()),
        (_("Operating System Version"), platform.version()),
        (_("Architecture"), " / ".join(platform.architecture())),
        (_("Processors Type"), platform.processor()),
        (
            _("Processors Quantity"),
            f"{psutil.cpu_count(logical=False)} / {psutil.cpu_count()}",
        ),
        (
            _("Processors Frequency"),
            f"{(psutil.cpu_freq().current / 1024):.1f} Ghz",
        ),
        (
            _("Processors Load"),
            f"{(psutil.getloadavg()[0] / psutil.cpu_count() * 100):.2f} %",
        ),
        (
            _("Total Memory"),
            f"{(psutil.virtual_memory().total / (pow(1024, 3))):.2f} Gb",
        ),
        (_("Memory Usage"), f"{(psutil.virtual_memory().percent):.2f} %"),
        (_("Python Version"), platform.python_version()),
    ]
    return render_template("system.html.jinja", system_info=system_info)


@app.route("/application.html")
@logger.catch
def application():
    logger.trace("application()")
    application_info = [
        (_("Version"), f"odoo-pos-version, {__version__}"),
        (_("Start datetime"), interface.start_date),
    ]
    return render_template("application.html.jinja", application_info=application_info)


@app.route("/photo_booth.html", methods=["GET"])
@logger.catch
def photo_booth():
    logger.trace("photo_booth()")
    return render_template("photo_booth.html.jinja")


@app.route("/errored_tasks.html/<string:device_type>")
@logger.catch
def errored_tasks(device_type):
    logger.trace("errored_tasks()")
    if device_type not in interface.device_types:
        raise Exception(f"{device_type} is not a valid device type.")
    device = interface.get_device(device_type)
    return render_template("errored_tasks.html.jinja", device=device)


@app.route("/disconnections.html/<string:device_type>")
@logger.catch
def disconnections(device_type):
    logger.trace("disconnections()")
    if device_type not in interface.device_types:
        raise Exception(f"{device_type} is not a valid device type.")
    device = interface.get_device(device_type)
    return render_template("disconnections.html.jinja", device=device)



_DISPLAY_DEVICES = {
    (0x0416, 0xF012): {
        "name": "Aures - OCD 300",
        "image": "display__aures__ocd_300.png",
    },
}

_PRINTER_DEVICES = {
    (0x04B8, 0x0E15): {
        "name": "Epson - TM-T20II",
        "image": "printer__epson__tm_t20.png",
    },
    (0x04B8, 0x0E28): {
        "name": "Epson - TM-T20III",
        "image": "printer__epson__tm_t20.png",
    },
}

_PAYMENT_DEVICES = {
    (0x079B, 0x0028): {
        "name": "Ingenico - Move/5000",
        "image": "payment__ingenico__move_5000.png",
    },
}

_SCALE_DEVICES = {
    (0x0EB8, 0x2200): {
        "name": "Mettler Toledo - Ariva S",
        "image": "scale__mettler_toledo__ariva_s.png",
    },
}


def get_device_type(vendor_code, product_code):
    """return the type of the device, depending on the
    vendor and product code;
    vendor_code: hexadecimal, example: 0x0416
    product_code: hexadecimal, example: 0xF012
    Return: value in ['display', 'printer', 'payment', 'scale', False]
    and extra data
    """
    vendor_product_code = (vendor_code, product_code)

    if vendor_product_code in _DISPLAY_DEVICES.keys():
        device_type = "display"
        extra_info = _DISPLAY_DEVICES.get(vendor_product_code)

    elif vendor_product_code in _PRINTER_DEVICES.keys():
        device_type = "printer"
        extra_info = _PRINTER_DEVICES.get(vendor_product_code)

    elif vendor_product_code in _PAYMENT_DEVICES.keys():
        device_type = "payment"
        extra_info = _PAYMENT_DEVICES.get(vendor_product_code)

    elif vendor_product_code in _SCALE_DEVICES.keys():
        device_type = "scale"
        extra_info = _SCALE_DEVICES.get(vendor_product_code)

    else:
        device_type = extra_info = False

    return (device_type, extra_info)



import random

from loguru import logger

from ..tools.serial_code import ENQ, EOT, ESC, ETX, STX
from .scale_answer import ScaleAnswer


# ##################################################
# Write Orders
# ##################################################
def send_info(device, unit_price, tare, extra_text):
    unit_price_txt = str(int(unit_price * 100)).rjust(6, "0")
    tare_txt = str(int(tare * 1000)).rjust(4, "0")
    extra_text_txt = extra_text.rjust(13, " ")

    message = [EOT, STX]

    if not tare and not extra_text:
        message += ["01", ESC, unit_price_txt, ESC, ETX]
        description = "Send Unit Price"

    elif not tare and extra_text:
        message += ["04", ESC, unit_price_txt, ESC, extra_text_txt, ETX]
        description = "Send Unit Price and Extra Text"

    elif tare and not extra_text:
        message += ["03", ESC, unit_price_txt, ESC, tare_txt, ETX]
        description = "Send Unit Price and Tare"

    else:
        message += ["05", ESC, unit_price_txt, ESC, tare_txt, ESC, extra_text_txt, ETX]
        description = "Send Unit Price, Tare and Extra Text"

    _send_message(device, message, description)

    return get_answer(device)


def send_checksums(device, polynomial, checksums_d0, checksums_z):
    checksums_cs, checksums_kw = _compute_cs_kw(polynomial, checksums_d0, checksums_z)
    _send_message(
        device,
        [EOT, STX, "10", ESC, checksums_cs, checksums_kw, ETX],
        "Send CS and KW computed values",
    )
    return get_answer(device)


def send_ask_weight(device):
    _send_message(device, [EOT, ENQ], "Ask weight")
    return get_answer(device)


def send_ask_status(device):
    _send_message(device, [EOT, STX, "08", ETX], "Ask Scale Status")
    return get_answer(device)


def _send_EOT(device):
    _send_message(device, [EOT], False)


def _send_message(device, message_list, action_description):
    message = b"".join([x.encode("utf-8") for x in message_list])
    if action_description:
        logger.debug(f"Scale: {action_description} with the message {message} ...")
    device.write(message)


# ##################################################
# Read Orders
# ##################################################


def get_answer(device):
    raw_answer = _get_raw_answer(device)
    answer = ScaleAnswer(raw_answer)
    logger.debug(f"Scale: interprated answer received: {answer.answer_type}.")
    return answer


def _get_raw_answer(device):
    char_list = []
    while True:
        char = device.read(1)  # may return `bytes` or `str`
        if not char:
            break
        else:
            char_list.append(bytes(char))
    result = b"".join(char_list)
    logger.debug(f"Scale: Raw answer received: {result}")
    _send_EOT(device)
    return result


# ##################################################
# Checksums Functions
# ##################################################


def _compute_cs_kw(polynomial, checksums_d0, checksums_z):
    cs = _generate_random_hex_code()
    cs_size = len(cs) * 4
    cs_bin = bin(int(cs, 16))[2:].zfill(cs_size)

    # apply first bit of z on checksum
    cs_bin_encoded = _rotate_left(cs_bin, int(checksums_z[0], 16))
    cs_hex_ascii_encoded = "{:04X}".format(int(cs_bin_encoded, 2))

    # generate kw
    kw = _get_kw_value(polynomial, int(cs, 16))
    kw_size = len(kw) * 4
    kw_bin = bin(int(kw, 16))[2:].zfill(kw_size)

    # apply second bit of z on kw
    kw_bin_encoded = _rotate_right(kw_bin, int(checksums_z[1], 16))
    # kw_hex_ascii_encoded = "{:04X}".format(int(kw_bin_encoded, 2), "X")
    kw_hex_ascii_encoded = "{:04X}".format(int(kw_bin_encoded, 2))

    return cs_hex_ascii_encoded, kw_hex_ascii_encoded


def _get_kw_value(udw_generator, uw_checksum):
    ub_shifts = 0
    if not udw_generator:
        return 0
    udw_kw = uw_checksum << 16
    while not (udw_generator & 0x80000000):
        udw_generator <<= 1
    udw_kw ^= udw_generator
    while not (udw_kw & 0x80000000):
        udw_kw <<= 1
        ub_shifts += 1
        if ub_shifts == 16:
            break
    while ub_shifts < 16:
        udw_kw ^= udw_generator
        while not (udw_kw & 0x80000000):
            udw_kw <<= 1
            ub_shifts += 1
            if ub_shifts == 16:
                break
    udw_kw >>= 16
    return format(udw_kw, "04x")


def _generate_random_hex_code():
    return "".join([random.choice("0123456789ABCDEF") for x in range(4)])


def _rotate_left(num, bits):
    """
    The encoding of the CS-values has to be made
    by rotating them to the left for n bits
    """
    debut = num[0:bits]
    fin = num[bits:16]
    return "".join(fin + debut)


def _rotate_right(num, bits):
    """
    The encoding of the KW-values has to be made
    by rotating them to the right for n bits
    """
    debut = num[16 - bits : 16]
    fin = num[0 : 16 - bits]
    return "".join(debut + fin)



from telium import Telium, TeliumAsk

from .models.device_abstract import DeviceAbstract


class DevicePayment(DeviceAbstract):
    device_type = "payment"

    def push_amount(self, amount, currency_iso):
        if not self.terminal_file:
            return False
        telium = Telium(self.terminal_file)
        payment = TeliumAsk.new_payment(
            amount,
            payment_mode="debit",
            target_currency=currency_iso,
        )
        result = telium.ask(payment)
        telium.close()
        return result



from flask import jsonify, request
from loguru import logger

from ..app import app
from ..interface import interface


@app.route("/hw_proxy/scale_read", methods=["POST", "PUT"])
@logger.catch
def scale_read():
    logger.trace("scale_read()")
    data = request.json.get("params", {}).get("data", {})
    interface.log_data(data, "scale", "DEBUG")

    try:
        unit_price = float(data.get("unit_price", 0.0))
    except ValueError:
        unit_price = 0.0
    try:
        tare = float(data.get("tare", 0.0))
    except ValueError:
        tare = 0.0

    result = interface.device_scale_read_weight(unit_price, tare)

    return jsonify(jsonrpc="2.0", result=result)



import re

from loguru import logger

from ..tools.serial_code import ACK, ESC, ETX, NAK, STX

error_regex__record_09 = re.compile(rf"{STX}09{ESC}(?P<error>\d+){ETX}")

checksums_request_regex__record_11 = re.compile(
    rf"{STX}11{ESC}(?P<checksums_d0>\d)(?P<checksums_z>\w\w){ETX}"
)
checksums_ok_regex__record_11 = re.compile(rf"{STX}11{ESC}1{ETX}")

weighing_result_regex__record_02 = re.compile(
    rf"{STX}02{ESC}(?P<status>\w){ESC}(?P<weight>\d+)"
    rf"{ESC}(?P<unit_price>\d+){ESC}(?P<total_price>\d+){ETX}"
)

_ERROR_DESCRIPTIONS = {
    "00": "no error",
    "01": "general error",
    "02": "parity error or buffer overflow",
    "10": "invalid record no.",
    "11": "invalid unit price",
    "12": "invalid tare value",
    "13": "invalid text",
    "20": "scale is still in motion (no standstill)",
    "21": "scale wasn't in motion since last operation",
    "22": "measurement is not yet finished",
    "30": "weight is less than minimum weight",
    "31": "scale is less than 0",
    "32": "scale is overloaded",
    "-1": "no response. Please reboot the scale",
    "-2": "unimplemented communication with scale",
}


class ScaleAnswer:
    answer_type = ""
    _byte_message = b""

    def __repr__(self):
        result = f"weight: {self.weight} - error_number: {self.error_number} - error_description: {self.error_description}"
        return result

    @property
    def error_description(self):
        if self.error_number:
            return _ERROR_DESCRIPTIONS.get(self.error_number, "Unknown Error")
        return ""

    def __init__(self, byte_answer):
        self.error_number = False
        self.weight = False

        self._byte_message = byte_answer
        self._message = byte_answer.decode("utf-8")

        if self._message == NAK:
            self.answer_type = "record_NAK"

        elif self._message == ACK:
            self.answer_type = "record_ACK"

        elif checksums_request_regex__record_11.search(self._message):
            self.answer_type = "record_11_checksums_request"
            result = checksums_request_regex__record_11.search(self._message)
            self.checksums_d0 = result.group("checksums_d0")
            self.checksums_z = result.group("checksums_z")

        elif checksums_ok_regex__record_11.search(self._message):
            self.answer_type = "record_11_checksums_valid"

        elif weighing_result_regex__record_02.search(self._message):
            self.answer_type = "record_02_weight"
            result = weighing_result_regex__record_02.search(self._message)
            self.weight = float(result.group("weight")) / 1000
            self.total_price = result.group("total_price")

        elif error_regex__record_09.search(self._message):
            self.error_number = error_regex__record_09.search(self._message).group(
                "error"
            )
            if self.error_number == "00":
                self.answer_type = "record_09_no_error"
            else:
                self.answer_type = "record_09_error"

        elif self._message == "":
            # In some very rare random cases, scale doesn't return nothing
            # In that case, we return an unknown error
            # Next communication will work correctly
            self.error_number = "-1"
            self.answer_type = "unknown_error"
            logger.error("No answer from scale")

        else:
            # This part should never occures
            self.error_number = "-2"
            self.answer_type = "unknown_error"
            logger.critical(f"Unimplemented Feature. Message: {self._byte_message}")



import base64
from io import BytesIO

from escpos.printer import Usb
from PIL import Image

from .models.device_abstract_with_queue import DeviceAbstractWithQueue


class DevicePrinter(DeviceAbstractWithQueue):
    device_type = "printer"

    def _process_task(self, task):
        printer = Usb(self._usb_device.idVendor, self._usb_device.idProduct, 0)
        (action, value) = task.data

        if action == "print":
            im = Image.open(BytesIO(base64.b64decode(value)))
            printer.image(im)
            printer.cut()

        elif action == "open_cashbox":
            printer.cashdraw(2)

        printer.close()
